# Bounce - a Dragonfly Benchmark

Mark Claypool, WPI<p>
`claypool@cs.wpi.edu`<p>

v3.1

## Overview

Bounce continually creates game objects (bouncing balls) until the
game engine can no longer keep up.  The balls move with random speeds
and directions and collide with other balls.  The balls are created at
a rate of 1 ball per frame (so, about 30 balls per second).  Bounce
runs three game sessions (starting with zero balls each time) and then
produces a final report of the average number (over the 3 runs) of
game objects Dragonfly can support under the tested conditions.

## How to Use

### Linux

0) Extract the benchmark and edit the Makefile.

1) Compile by typing:

> make depend

then:

> make clean

then:
  
> make

2) Run by typing:
  
> bounce

3) When done, the benchmark reports the maximum number of objects on
the screen.

### Windows (Visual Studio 2013):

0) Extract the benchmark and open the Visual Studio solution in
directory vs-2013/.

1) Compile (hotkey F7).

2) Run (hotkey F5).
  
3) When done, dragonfly.log reports the maximum number of objects.

## Details

+ Initially there are zero Balls.

+ One Ball is created per step (about 30/second).

+ Each ball has a random speed (from 0.1 to 1.0 spaces/step) and
  direction.

+ Balls are solid and collide with other objects and the view
  boundary.

+ The frame time (i.e., how long the game loop took) is recorded for
  the latest 30 steps (about 1 second).

+ If the median frame time is 10% over the target frame time (33 ms),
  then the session is done.

+ When an session is complete, the number of Balls is recorded.

+ After three sessions are run, the number of Balls over the three
  sessions is averaged and reported.
  
+ This average is called the *bouncemark* and provides an estimate of
  the maximum number of objects the setup can support.


## Uses

Overall, the bouncemark (B) gives a reasonable maximum for the number
of moving objects the game engine can support under the tested
environment (this means for that specific computer, operating system,
and display device).

If there is just a single moving game object, Dragonfly can support
about B times as many objects, so B*B (e.g. a game with a moving Hero
and lots of Walls).

Note, however, that any game-code computations will reduce the number
of game objects that can be supported since bounce has minimal game
code.

In general:

````
   B = estimated maximum reported by Bounce
   M = number of moving objects
   S = number of static (non-moving) objects
````

For "good" game performance, ensure:

````
    M * (M + S) <= (B*B)
````

## Notes

+ The Bounce benchmark uses a sliding window to keep track of
performance (the frame times).  This provides a tradeoff between an
instantaneous value (using just the most recent frame time) and a
continuous value (using all the frame times from the start).

+ For the frame time statistic, a median is used to avoid spurious
outliers that may not reflect the true, sustained load the setup can
handle.

+ Alternatives to a windowed average, such as an exponentially
weighted moving average (EWMA), were considered, but even an EWMA with
a low alpha was found to still be susceptible to a single, large event
outlier.

+ The 10% threshold for performance degradation was chosen based on
many studies that suggested this value provided the true start to the
climb in game loop times as game objects continued to be added.

## Future

Future additions may attempt to incorporate other aspects of games, as
well as features of Dragonfly, including but not limited to:

+ Pathfinding.  Perhaps 1 in 30 objects pathfinds then travels to a
random location, continually.

+ Input.  Have one object get input at a typical game rate (e.g., 2 or
3 per second).

+ Game world.  Make bounce work normally, but have the game world
larger than the view world.  Maybe 9x larger.

+ Sprites. Have some objects draw (and animate) multi-character
sprites of various sizes.
